/* Copyright 2020, Michele Santullo
 * This file is part of user-gcc.
 *
 * User-gcc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * User-gcc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with user-gcc.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "list_profiles.hpp"
#include "syspaths.hpp"
#include "config.h"
#include "chost.hpp"
#include "to_var_map.hpp"
#include "load_file.hpp"
#include "string_view_cat.hpp"
#include "gentoofunctions.hpp"
#include <iostream>
#include <filesystem>
#include <vector>
#include <string>
#include <algorithm>
#include <regex>

namespace duck {
	namespace fs = std::filesystem;

	namespace {
		[[gnu::pure]]
		bool starts_with (std::string_view s, std::string_view prefix) {
			return (s.substr(0, std::min(prefix.size(), s.size())) == prefix);
		}

		std::vector<int> decompose_version (const std::string& ver) {
			std::vector<int> ret;
			int curr = 0;
			bool push_last = false;
			for (char c : ver) {
				if (c >= '0' and c <= '9') {
					curr = (curr * 10) + (c - '0');
					push_last = true;
				}
				else if (push_last) {
					ret.push_back(curr);
					curr = 0;
					push_last = false;
				}
			}
			if (push_last)
				ret.push_back(curr);
			return ret;
		}

		[[gnu::pure]]
		bool version_comp (const std::string& a, const std::string& b) {
			std::vector<int> va(decompose_version(a));
			std::vector<int> vb(decompose_version(b));
			const std::size_t sz = std::max(va.size(), vb.size());
			va.resize(sz, 0);
			vb.resize(sz, 0);

			for (std::size_t z = 0; z < sz; ++z) {
				if (va[z] < vb[z])
					return true;
				else if (va[z] > vb[z])
					return false;
			}
			return false;
		}

		std::string current_compiler (const fs::path& config_path, const std::string& ctarget) {
			if (fs::is_regular_file(config_path)) {
				auto text = load_file(config_path);
				auto var_map = to_var_map(text);
				auto found_it = var_map.find("CURRENT");
				if (var_map.end() == found_it)
					return {};
				else
					return to_string(found_it->second);
			}
			else {
				return {};
			}
		}
	} //unnamed namespace

	std::vector<fs::path> version_sorted_paths (const fs::path path) {
		std::vector<fs::path> ret;
		for (const fs::directory_entry& curr : fs::directory_iterator(path)) {
			std::string filename = curr.path().filename();
			if (not starts_with(filename, "config") and curr.is_regular_file())
				ret.push_back(curr);
		}
		std::regex version_reg(
			R"(-(\d+(?:\.\d)*(?:r\d+)?))",
			std::regex_constants::ECMAScript | std::regex_constants::optimize
		);
		std::sort(
			ret.begin(),
			ret.end(),
			[&version_reg](const fs::path& a, const fs::path& b) {
				using std::regex_match;
				std::smatch ma, mb;
				std::string sa = a.filename();
				std::string sb = b.filename();
				return version_comp(
					(regex_search(sa, ma, version_reg) ? ma[1] : std::string("0")),
					(regex_search(sb, mb, version_reg) ? mb[1] : std::string("0"))
				);
			}
		);
		return ret;
	}

	void list_profiles(const SysPaths& sp) {
		using std::string;

		if (sp.root() != "/") {
			//echo "Using gcc-config info in ${ROOT}"
			std::cout << "Using " << PROJECT_NAME << " info in " << sp.root() << '\n';
		}

		std::string chost = duck::chost(sp);
		const std::string& ctarget = chost;

		std::string_view filter;
		auto config_path = sp.gcc_env_d() / ("config-" + ctarget);
		auto s = fs::status(config_path);
		if (not fs::is_regular_file(s)) {
			filter = ctarget;
		}

		const std::string current = current_compiler(config_path, ctarget);
		std::string_view current_native(current);
		auto version_paths = version_sorted_paths(sp.gcc_env_d());
		std::string text;
		std::size_t z = 0;
		std::string target;
		for (const auto& ver_path : version_paths) {
			text = load_file(ver_path);
			auto var_map = to_var_map(text);
			auto curr_ctarget = var_map["CTARGET"];
			std::string filename = ver_path.filename();

			if (curr_ctarget.empty()) {
				if (starts_with(filename, chost + "-"))
					curr_ctarget = chost;
				else
					ewarn(sp, "broken config file: " + std::string(ver_path));
			}

			++z;
			if (not filter.empty() and filter != curr_ctarget)
				continue;
			if (target != curr_ctarget) {
				if (z > 1)
					std::cout << '\n';
				target = to_string(curr_ctarget);
				curr_ctarget = "";
			}

			std::cout << " [" << z << "] " << filename;
			auto colours = sp.colours();
			if (filename == current_native) {
				std::cout << ' ' << colours["GOOD"] << '*' << colours["NORMAL"];
			}
			else if (fs::is_regular_file(sp.gcc_env_d() / ("config-" + target)) and filename == current) {
				std::cout << ' ' << colours["HILITE"] << '*' << colours["NORMAL"];
			}
			std::cout << '\n';
		}
	}
} //namespace duck
