/* Copyright 2020, Michele Santullo
 * This file is part of user-gcc.
 *
 * User-gcc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * User-gcc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with user-gcc.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "syspaths.hpp"
#include "to_var_map.hpp"
#include "string_view_cat.hpp"
#include <unistd.h>
#include <cassert>

namespace duck {
	namespace {
		std::string fix_eprefix(std::string_view p) {
			if (p.size() >= 2 and p.front() == '@' and p.back() == '@')
				return "";
			else
				return std::string(p);
		}

		std::string_view remove_front_slash(std::string_view text) {
			const std::size_t pos = text.find_first_not_of('/');
			return (text.npos == pos ? text : text.substr(pos));
		}

		std::string_view val_or_def (const VarMap& map, std::string_view key, std::string_view def) noexcept {
			try {
				auto itm = map.find(key);
				return (map.end() == itm ? def : itm->second);
			}
			catch (...) {
				return def;
			}
		}
	} //unnamed namespace

	SysPaths::SysPaths() :
		m_env(to_var_map_ignore_ownership(environ)),
		m_root(val_or_def(m_env, "ROOT", "/")),
		m_eprefix(fix_eprefix(val_or_def(m_env, "EPREFIX", ""))),
		m_eroot(m_root / remove_front_slash(std::string(m_eprefix))),
		m_env_d(m_eroot / "etc/env.d"),
		m_gcc_env_d(m_env_d / "gcc")
	{
	}

	std::string_view SysPaths::operator[] (std::string_view name) const noexcept {
		assert(not name.empty());
		return val_or_def(m_env, name, "");
	}

	std::optional<std::string_view> SysPaths::env (std::string_view name) const noexcept {
		try {
			assert(not name.empty());
			auto it_found = m_env.find(name);
			if (m_env.end() == it_found)
				return {};
			else
				return std::make_optional(it_found->second);
		}
		catch (...) {
			return {};
		}
	}

	std::string_view SysPaths::env (std::string_view name, std::string_view def) const noexcept {
		assert(not name.empty());
		return val_or_def(m_env, name, def);
	}

	void SysPaths::set_colours (std::string_view colours) {
		m_colours = to_var_map_ignore_ownership(colours);
	}

	auto SysPaths::root() const -> const path_type& {
		return m_root;
	}

	auto SysPaths::eroot() const -> const path_type& {
		return m_eroot;
	}

	auto SysPaths::eprefix() const -> const path_type& {
		return m_eprefix;
	}

	auto SysPaths::env_d() const -> const path_type& {
		return m_env_d;
	}

	auto SysPaths::gcc_env_d() const -> const path_type& {
		return m_gcc_env_d;
	}

#if !defined(NDEBUG)
	std::ostream& operator<< (std::ostream& oss, const SysPaths& sp) {
		oss << "SysPaths {\n" <<
			"\tROOT: " << sp.root() << '\n' <<
			"\tEPREFIX: " << sp.eprefix() << '\n' <<
			"\tEROOT: " << sp.eroot() << '\n' <<
			"\tENV_D: " << sp.env_d() << '\n' <<
			"\tGCC_ENV_D: " << sp.gcc_env_d() << '\n' <<
			'}';
		return oss;
	}
#endif
} //namespace duck
